import mt.MLib;
import mt.deepnight.slb.*;
import mt.deepnight.Lib;
import flash.display.Bitmap;
import flash.display.BitmapData;

class Entity {
	public static var ALL : Array<Entity> = [];

	public var spr				: BSprite;
	public var destroyAsked		: Bool;
	public var cd				: mt.Cooldown;

	public var cx				: Int;
	public var cy				: Int;
	public var xr				: Float;
	public var yr				: Float;
	public var xx(get,never)	: Float;
	public var yy(get,never)	: Float;
	public var dx				: Float;
	public var dy				: Float;
	public var dir				: Int;

	var scene(get,never)		: Scene;
	public var stable(get,never): Bool;
	var time(get,never)			: Int;
	public var crouch			: Bool;
	public var life				: Int;
	public var maxLife			: Int;

	public var lifeBar			: flash.display.Bitmap;

	public var tx				: Null<Float>;

	public function new() {
		cd = new mt.Cooldown();
		destroyAsked = false;
		ALL.push(this);
		setPosCase(0,0);
		dx = dy = 0;
		crouch = false;
		dir = 1;

		spr = new BSprite( Assets.tiles );
		Game.ME.sdm.add(spr, Const.DP_HERO);
		spr.setCenter(0.5,1);

		lifeBar = new flash.display.Bitmap(new flash.display.BitmapData(32,5,true,0x0));
		Game.ME.sdm.add(lifeBar, Const.DP_HERO);
		lifeBar.filters = [
			new flash.filters.DropShadowFilter(1,-90, 0x0,0.4, 0,0,1, 1,true),
			new flash.filters.GlowFilter(0x0,1, 2,2, 10),
		];

		initLife(3);
	}

	public inline function setCrouch(v) crouch = v;

	public inline function isWalking() return MLib.fabs(dx)>0.03;

	public function initLife(n) {
		life = maxLife = n;
		updateLifeBar();
	}

	public function hit() {
		life--;
		updateLifeBar();
		if( life<=0 && !cd.has("death") )
			onDie();
	}

	public inline function isDead() return life<=0;

	function onDie() {
		cd.set("death", Const.seconds(3));
		Fx.ME.dead(this);
	}

	function updateLifeBar() {
		var bd = lifeBar.bitmapData;
		bd.fillRect(bd.rect, mt.deepnight.Color.addAlphaF(0));
		var w = 2;
		var h = 5;
		bd.fillRect(new flash.geom.Rectangle(0,0, maxLife*(w+1),h), mt.deepnight.Color.addAlphaF(0x33394A));
		for(i in 0...maxLife) {
			var r = new flash.geom.Rectangle(i*(w+1),0,w,h);
			bd.fillRect(r, mt.deepnight.Color.addAlphaF(0x262B37));
		}
		for(i in 0...life) {
			var r = new flash.geom.Rectangle(i*(w+1),0,w,h);
			bd.fillRect(r, mt.deepnight.Color.addAlphaF(life==1 ? 0xFF0000 : life==2 ? 0xFF9300 : 0xBFFF00));
		}
	}

	inline function get_scene() return Scene.CURRENT;
	inline function get_time() return Game.ME.time;
	inline function rnd(min,max,?sign) return Lib.rnd(min,max,sign);
	inline function irnd(min,max,?sign) return Lib.irnd(min,max,sign);

	public function setPosCase(x,y, ?d=0) {
		cx = x;
		cy = y;
		xr = 0.5;
		yr = 1;
		dx = dy = 0;
		if( d!=0 )
			dir = d;
	}

	public function setPosFree(x,y, ?d=0) {
		cx = Std.int(x/Const.GRID);
		cy = Std.int(y/Const.GRID);
		xr = (x-cx*Const.GRID) / Const.GRID;
		yr = 1;
		//yr = (y-cy*Const.GRID) / Const.GRID;
		if( d!=0 )
			dir = d;
	}

	public function destroy() {
		destroyAsked = true;
	}

	public function unregister() {
		ALL.remove(this);

		lifeBar.parent.removeChild(lifeBar);
		lifeBar.bitmapData.dispose();
		lifeBar.bitmapData = null;
		lifeBar = null;

		cd.destroy();
		cd = null;

		spr.dispose();
		spr = null;
	}


	public function onEndFrame() {
		spr.x = Std.int(xx);
		spr.y = Std.int(yy);
		spr.scaleX = MLib.fabs(spr.scaleX) * dir;

		lifeBar.visible = spr.visible && Game.ME.hero.combat && !scene.locked;
		lifeBar.x = Std.int( spr.x-maxLife*3*0.5 );
		lifeBar.y = Std.int( spr.y+10 );
	}


	public inline function goto(x) {
		crouch = false;
		tx = x;
	}


	inline function get_xx() return (cx+xr)*Const.GRID;
	inline function get_yy() return (cy+yr)*Const.GRID;
	inline function get_stable() return yr==1 && scene!=null && scene.hasCollision(cx,cy+1);

	public function update() {
		cd.update();

		if( tx!=null ) {
			dir = (tx>xx ? 1 : -1);
			dx += 0.03 * dir;
		}
		if( MLib.fabs(xx-tx)<=5 ) {
			scene.cm.signal("goto");
			tx = null;
		}

		// X
		xr+=dx;
		dx*=0.7;
		if( xr>=0.8 && scene.hasCollision(cx+1,cy) ) {
			xr = 1;
			dx = 0.8;
		}
		if( xr>=0.5 && scene.hasCollision(cx+1,cy) ) {
			if( dx>0 )
				dx = 0;
			xr-=0.04;
		}
		if( xr<=0.2 && scene.hasCollision(cx-1,cy) ) {
			xr = 0;
			dx = 0.2;
		}
		if( xr<0.5 && scene.hasCollision(cx-1,cy) ) {
			if( dx<0 )
				dx = 0;
			xr+=0.04;
		}
		while( xr>1 ) {
			xr--;
			cx++;
		}
		while( xr<0 ) {
			xr++;
			cx--;
		}


		// Y
		yr+=dy;
		dy*=0.8;
		if( yr<1 || !scene.hasCollision(cx,cy+1) )
			dy+=0.04;
		if( yr>=1 && scene.hasCollision(cx,cy+1) ) {
			yr = 1;
			dy = 0;
		}
		if( yr<1 && scene.hasCollision(cx,cy-1) ) {
			yr = 1;
			dy = 0;
		}
		while( yr>1 ) {
			yr--;
			cy++;
		}
		while( yr<0 ) {
			yr++;
			cy--;
		}

	}
}