import mt.deepnight.Buffer;
import flash.display.Sprite;
import mt.MLib;
import mt.flash.Key;
import flash.ui.Keyboard;
import mt.deepnight.Lib;
import mt.deepnight.slb.*;

class Game extends mt.deepnight.FProcess { //}
	public static var ME : Game;

	public var buffer		: Buffer;
	public var scroller		: flash.display.Sprite;
	public var sdm			: mt.flash.DepthManager;
	public var fx			: Fx;
	public var hero			: en.Hero;
	public var viewport		: { x:Float, y:Float };
	var poiLabel			: flash.text.TextField;
	var line				: flash.display.Sprite;
	public var dark			: BSprite;
	var scene(get,never)	: Scene;

	public function new() {
		super();
		ME = this;

		buffer = new Buffer(350,150, Const.UPSCALE, false, 0x0);
		root.addChild(buffer.render);

		scroller = new flash.display.Sprite();
		buffer.dm.add( scroller, Const.DP_BG );
		buffer.setTexture( Buffer.makeGrid(Const.UPSCALE, Const.UPSCALE, 0x3F64BA), 0.5, true );
		sdm = new mt.flash.DepthManager(scroller);

		fx = new Fx();

		hero = new en.Hero();

		viewport = { x:hero.xx, y:hero.yy-50 }

		line = new flash.display.Sprite();
		Game.ME.sdm.add(line, Const.DP_UI);
		line.graphics.lineStyle(1, 0x8ADBEC, 0.8, true);
		line.graphics.moveTo(0,0);
		line.graphics.lineTo(16,0);
		line.blendMode = ADD;
		line.filters = [
			new flash.filters.GlowFilter(0x187C92,0.5, 4,4,5),
		];

		poiLabel = Assets.createField("???");
		Game.ME.sdm.add(poiLabel, Const.DP_UI);
		poiLabel.filters = [
			new flash.filters.GlowFilter(0x187C92,1, 2,2,10),
		];
		setPoiLabel();

		dark = Assets.tiles.get("dark");
		buffer.dm.add(dark, Const.DP_DARK);
		dark.width = buffer.width;
		dark.height = buffer.height;
		dark.alpha = 0.8;

		#if debug
		ui.TutorialMsg.complete("use");
		ui.TutorialMsg.complete("move");
		Assets.music.playLoop();
		#end

		#if !debug
		new sc.Balcony();
		#else
		//new en.Cop(15,8);
		//new sc.Balcony(); hero.setPosCase(29,8);
		new sc.Apartment();
		//new sc.Bathroom();
		#end

		flash.Lib.current.stage.addEventListener( flash.events.MouseEvent.CLICK, onClick );
		tw.create(root.alpha, 0>1, 1500);
	}

	inline function get_scene() return Scene.CURRENT;

	function onClick(_) {
		#if debug
		if( mt.flash.Key.isDown(mt.flash.Key.DELETE) ) {
			var m = getMouse();
			flash.system.System.setClipboard(m.sx+","+m.sy);
		}
		#end
	}

	public function getMouse() {
		var s = flash.Lib.current.stage;
		return {
			bx	: Std.int(s.mouseX/Const.UPSCALE),
			by	: Std.int(s.mouseY/Const.UPSCALE),

			sx	: Std.int(s.mouseX/Const.UPSCALE - scroller.x),
			sy	: Std.int(s.mouseY/Const.UPSCALE - scroller.y),
		}
	}


	public function setPoiLabel(?p:Poi) {
		if( p!=null && p.name==poiLabel.text )
			return;


		poiLabel.visible = p!=null;
		line.visible = poiLabel.visible;
		if( p!=null ) {
			poiLabel.alpha = 0;
			poiLabel.text = p.name;
			poiLabel.width = 200;
			poiLabel.x = 5 + Std.int( p.x + p.wid*0.5 - poiLabel.textWidth*0.5 );
			poiLabel.y = Std.int( p.y - poiLabel.textHeight - 16 );

			line.x = poiLabel.x + poiLabel.textWidth*0.5;
			line.y = poiLabel.y + poiLabel.textHeight + 2;
			line.rotation = MLib.toDeg( Math.atan2(p.centerY-line.y, p.centerX-line.x) );

			tw.create(poiLabel.scaleY, 0>1, 150);
			tw.create(line.scaleX, 0>1, 300);
		}
		else
			poiLabel.text="";
	}

	public function getClosestPoi(x,y) {
		var all = Scene.CURRENT.poi.filter( function(p) return p.isClose() );
		var best : Poi = null;
		for( p in all )
			if( best==null || Lib.distanceSqr(p.centerX, p.centerY, x,y) < Lib.distanceSqr(best.centerX, best.centerY, x,y) )
				best = p;

		if( best!=null && Lib.distanceSqr(best.centerX, best.centerY, x,y)>=40*40 )
			best = null;

		return best;
	}



	public function activate(x,y) {
		if( scene.locked || !ui.TutorialMsg.hasDone("use") )
			return;

		var p = getClosestPoi(x,y);
		if( p!=null ) {
			ui.TutorialMsg.clear("use");
			if( p.onUse!=null && p.onLook==null )
				p.onUse();
			else {
				var m = new ui.Menu(p);
				if( p.onLook!=null )
					m.addButton("Look", function() p.onLook() );

				if( p.onUse!=null )
					m.addButton("Use", function() p.onUse() );

				if( p.onUse!=null )
					m.addButton("Combine with...", function() {
						m.close();
						var inv = scene.getInventory();
						if( inv.length==0 )
							say("Your inventory is empty");
						else {
							var m = new ui.Menu(p,true);
							for(k in inv)
								m.addButton(k, function() {
									if( p.onUseWith==null || !p.onUseWith(k) )
										say("!This doesn't work.");
								});
						}
					});
			}
		}
	}

	function onAdded(_) {
		root.removeEventListener(flash.events.Event.ADDED, onAdded);

	}

	public function say(?str:String, ?a:Array<String>) {
		if( str!=null )
			a = [str];

		if( a.length!=0 ) {
			var p = new ui.Pop( a.shift() );
			if( a.length>0 )
				p.onClose = function() say(a);
		}
	}


	override function unregister() {
		super.unregister();


		while( Entity.ALL.length>0 ) {
			Entity.ALL[0].destroy();
			Entity.ALL[0].unregister();
		}

		mt.deepnight.Particle.clearAll();

		hero = null;
	}

	override function update() {
		super.update();

		// Entities
		for(e in Entity.ALL)
			e.update();
		for(e in Entity.ALL)
			e.onEndFrame();
		var i = 0;
		while( i<Entity.ALL.length )
			if( Entity.ALL[i].destroyAsked )
				Entity.ALL[i].unregister();
			else
				i++;

		// Point of interest
		var p = getClosestPoi(hero.xx, hero.yy);
		if( ui.TutorialMsg.hasDone("use") && !hero.combat && p!=null && ui.Pop.ALL.length==0 ) {
			fx.shinePoi(p);
			setPoiLabel( hero.isPressingMoveKeys() ? null : p);
			if( poiLabel.alpha<1 ) {
				poiLabel.alpha = MLib.fclamp(poiLabel.alpha+0.06, 0,1);
				line.alpha = poiLabel.alpha;
			}
		}
		else
			setPoiLabel();

		// General controls
		if( Key.isToggled(Keyboard.M) )
			mt.flash.Sfx.toggleMuteChannel(1);

		// Scrolling
		var tx = hero.xx + (scene.locked ? 0 : hero.dir*40);
		var ty = hero.yy-40;
		if( cd.has("credits") )
			tx = 0;
		if( hero.combat && en.Cop.ALL.length>0 ) {
			var e = en.Cop.ALL[0];
			tx = hero.xx + (e.xx-hero.xx)*0.5;
			ty = hero.yy + (e.yy-hero.yy)*0.5 - 40;
		}
		//tx = 300;
		var w = buffer.width;
		var h = buffer.height;
		if( Lib.distanceSqr(tx,ty, viewport.x, viewport.y)>=20*20 ) {
			viewport.x += (tx-viewport.x)*0.04;
			viewport.y += (ty-viewport.y)*0.04;
		}
		viewport.x = MLib.fclamp(viewport.x, w*0.5, scene.wid-w*0.5);
		viewport.y = MLib.fclamp(viewport.y, h*0.5, scene.hei-h*0.5);
		scroller.x = Std.int( -viewport.x+w*0.5 );
		scroller.y = Std.int( -viewport.y+h*0.5 );

	}
	override function render() {
		super.render();
		Assets.tiles.updateChildren();
		buffer.update();
	}
}
