class Poi {
	public var name		: String;
	public var x		: Float;
	public var y		: Float;
	public var wid		: Int;
	public var hei		: Int;
	var debug			: flash.display.Sprite;

	public var centerX(get,null)	: Float;
	public var centerY(get,null)	: Float;

	public var onUseWith: Null<String->Bool>;
	public var onUse	: Null<Void->Void>;
	public var onLook	: Null<Void->Void>;

	public function new(n:String, x:Float,y:Float, ?w=0., ?h=0.) {
		name = n;
		if( h==0 ) {
			// Circle
			if( w==0 )
				w = mt.MLib.ceil(Const.GRID);
			x -= w*0.5;
			y -= w*0.5;
			h = w;
		}
		this.x = x;
		this.y = y;
		this.wid = Std.int(w);
		this.hei = Std.int(h);

		debug = new flash.display.Sprite();
		Game.ME.sdm.add(debug, Const.DP_UI);
		debug.graphics.beginFill(0x00FF00,0.4);
		debug.graphics.drawRect(0,0,wid,hei);
		debug.x = x;
		debug.y = y;
		debug.visible = false;
		var tf = Assets.createField(name);
		debug.addChild(tf);

		Scene.CURRENT.poi.push(this);
	}

	public function isClose() {
		if( Scene.CURRENT.locked )
			return false;

		var h = Game.ME.hero;
		var r = new flash.geom.Rectangle(h.xx-7, h.yy-32, 14, 32);
		return r.intersects(new flash.geom.Rectangle(x,y,wid,hei));
	}

	public function snapshot() {
		var p = mt.MLib.max(16, Std.int((100-wid)*0.5));
		var bd = new flash.display.BitmapData(wid+p*2, hei+p*2, false, 0x0);
		var m = new flash.geom.Matrix();
		m.translate(p-x, p-y);
		Game.ME.setPoiLabel();
		bd.draw(Game.ME.scroller, m);
		return bd;
	}

	inline function get_centerX() return x + wid*0.5;
	inline function get_centerY() return y + hei*0.5;

	public function toString() return '$name@$x,$y,$wid,$hei';

	public function look(cb:Void->Void) {
		onLook = cb;
		return this;
	}

	public function use(cb:Void->Void) {
		onUse = cb;
		return this;
	}

	public function useWith(cb:String->Bool) {
		onUseWith = cb;
		return this;
	}

	public inline function show() debug.visible = true;
	public inline function hide() debug.visible = false;
}

