import mt.deepnight.slb.*;
import mt.deepnight.Color;
import mt.deepnight.Lib;
import mt.MLib;
import mt.flash.Key;

class Scene extends mt.deepnight.Process {
	public static var CURRENT : Scene;

	public static var SAVE : Map<String,Int> = null;
	public static var FLAGS : Map<String,Int> = new Map();
	public static var PREV_SCENE = "";

	public var map		: Array<Array<Bool>>;
	public var cwid		: Int;
	public var chei		: Int;
	public var bg		: BSprite;
	public var debug	: flash.display.Sprite;
	public var wid(get,never)	: Float;
	public var hei(get,never)	: Float;
	public var hero(get,never)	: en.Hero;
	public var g(get,never)		: Game;

	public var poi				: Array<Poi>;

	var coords					: flash.text.TextField;
	public var id				: String;
	public var cm				: mt.deepnight.Cinematic;
	public var locked			: Bool;


	public function new() {
		super(Game.ME);

		save();

		CURRENT = this;
		id = "??";
		locked = false;
		cm = new mt.deepnight.Cinematic(30);

		cwid = 30;
		chei = 10;
		poi = [];

		debug = new flash.display.Sprite();
		Game.ME.sdm.add(debug, Const.DP_UI);
		debug.visible = false;

		coords = Assets.createField("xxxx,yyyy", 0xFFFF00);
		Game.ME.sdm.add(coords, Const.DP_UI);
		coords.width = 150;
		coords.visible = false;
	}

	function removePoi(n:String) {
		for( e in poi.copy() )
			if( e.name==n )
				poi.remove(e);
	}

	public static function save() {
		SAVE = new Map();
		for(k in FLAGS.keys())
			SAVE.set(k, FLAGS.get(k));
	}

	public static function reload() {
		if( CURRENT!=null && SAVE!=null ) {
			var id = CURRENT.id;
			CURRENT.id = PREV_SCENE; // to trigger cameFrom events
			FLAGS = new Map();
			for(k in SAVE.keys())
				FLAGS.set(k, SAVE.get(k));

			CURRENT.transition( function() {
				var h = Game.ME.hero;
				h.spr.visible = true;
				h.setCombat(false);

				switch( id ) {
					case "balcony" : new sc.Balcony();
					case "apartment" : new sc.Apartment();
					case "bathroom" : new sc.Bathroom();
					default :
						trace("couldn't reload in "+id);
				}
			});
		}
	}

	public function onCombatEnd() {
	}

	function cameFrom(id:String) return PREV_SCENE==id;

	public static function resetSave() {
		FLAGS = new Map();
	}

	inline function get_g() return Game.ME;
	inline function get_hero() return Game.ME.hero;
	inline function get_wid() return cwid*Const.GRID;
	inline function get_hei() return chei*Const.GRID;

	public function set(k:String, ?v=1) {
		FLAGS.set(k,v);
		onStateChange(k);
	}
	public function unset(k:String) {
		FLAGS.remove(k);
		onStateChange(k);
	}
	public function has(k:String) return get(k)!=0;
	public function hasSet(k:String) {
		if( !has(k) ) {
			set(k);
			return false;
		}
		else
			return true;
	}
	public function get(k:String) return FLAGS.exists(k) ? FLAGS.get(k) : 0;
	public function toggle(k:String) {
		if( has(k) )
			set(k,0);
		else
			set(k,1);
		return has(k);
	}

	public function foundItem(k:String) return has("found_"+k);
	public function hasItem(k:String) return has("item_"+k);
	public function removeItem(k:String) set("item_"+k, 0);
	public function addItem(k:String) {
		set("found_"+k);
		set("item_"+k);
		Assets.SBANK.item(0.35);
		new ui.Pop("§New item: "+k, 20,60);
	}

	public function getInventory() {
		var all = [];
		for(k in FLAGS.keys())
			if( has(k) && k.indexOf("item_")==0 )
				all.push( k.substr(5) );
		return all;
	}

	function onStateChange(?k:String) {
	}

	public function objectiveComplete() {
		//#if debug
		//return true;
		//#else
		return has("sabotageFridge") && has("sabotageDoor");
		//#end
	}

	function transition(onDone:Void->Void) {
		locked = true;
		PREV_SCENE = id;
		ui.Pop.closeAll(false);
		ui.TutorialMsg.clearAll();

		if( ui.Menu.CURRENT!=null )
			ui.Menu.CURRENT.close();

		Game.ME.tw.create( Game.ME.root.alpha, 0 ).onEnd = function() {
			en.Cop.clear();
			mt.deepnight.Particle.clearAll();
			onDone();
			destroy();
			Game.ME.tw.create( Game.ME.root.alpha, 1 );
		}
	}

	public function read(id:String) {
		var bd = Assets.tiles.getBitmapData(id);
		cwid = bd.width;
		chei = bd.height;

		map = [];
		for(x in 0...cwid) {
			map[x] = [];
			for(y in 0...chei)
				map[x][y] = bd.getPixel(x,y)!=0x0;
		}

		bd.dispose();

		draw();
		onStateChange();
	}

	public function draw() {
		debug.graphics.clear();

		for(cx in 0...cwid)
			for(cy in 0...chei) {
				var x = cx*Const.GRID;
				var y = cy*Const.GRID;
				if( hasCollision(cx,cy) ) {
					debug.graphics.beginFill(0xFF00FF, 0.4);
					debug.graphics.drawRect(x,y,Const.GRID,Const.GRID);
				}
			}
	}

	public inline function isValid(x:Int,y:Int) {
		return x>=0 && x<cwid && y>=0 && y<chei;
	}

	public inline function hasCollision(x,y) {
		return isValid(x,y) ? map[x][y] : true;
	}


	override function unregister() {
		super.unregister();
		map = null;

		debug.parent.removeChild(debug);
		debug = null;

		bg.dispose();
		bg = null;

		if( CURRENT==this )
			CURRENT = null;
	}

	override function update() {
		super.update();

		cm.update();

		if( Key.isDown(Key.DELETE) ) {
			#if debug
			coords.visible = debug.visible = true;
			for(p in poi)
				p.show();
			var m = Game.ME.getMouse();
			coords.text = m.sx+","+m.sy+" hero="+hero.cx+","+hero.cy;
			coords.x = -Game.ME.scroller.x;
			coords.y = -Game.ME.scroller.y;
			#end
		}
		else {
			coords.visible = debug.visible = false;
			for(p in poi)
				p.hide();
		}
	}
}
